#!/bin/sh
# \
exec fltkwish "$0" -- ${1+"$@"}
#
# --- stnplot.tcl --- Station plotting application
#
# Copyright(C) I.B.Findleton, 2002. All Rights Reserved.
#
# This script monitors the observed weather data available on the web site
# of the Canadian Weather Service's main public site. The data for the
# selected list of stations is plotted as a standard station plot or is
# displayed as a weather depiction image. The data for the station is
# automatically updated by polling of the web site.
#
# The format of the command line call to this script is:
#
#	stnplot.tcl -s stn,stn,... -v verbose -m mode -w widgets -?
#				-h history -g gmt -d database -u units
#
# Where the stn values are station identifiers and the verbose parameter is
# either yes or no. The mode value can be plot or depiction. Any number of 
# station identifiers can be provided. The script will wrap the station 
# plots into a scrolled window if there are more than 4 stations specified.
#
# The mode specifies the default display mode. If it is plot, then something
# like the standard station plot model is used. If it is not plot, then an
# image depicting current conditions will be used.
#
# The widgets value is the number of widgets to show in the display window
# without having to scroll the list. Any number of stations can be displayed,
# but only this number will be visible at any one time. This is used to 
# conserve window real estate on your desktop.  
#
# Notes:
#	1 Canada is a metric country, so all data is metric unless you
#	   specify the -units imperial option
#	2 By default, the time values are GMT. Use -g 0 for local time
#	3 You need to use the station aviation identifiers (i.e YUL, KMIA)
#	   for the stations. Canadian stations need 3 letters, US stations
#	   have 4 letters, the first of which is K. You can also use the C
#	   for the Canadian stations, if you are a strict purist!
#	4 If you are on Linux/Unix and have the Tclgdbm extension, the data
#	   you download will be saved in a convenient database. On windows,
#	   you will need to roll your own archive by modifying this script.
#

# Check for a package needed by this script

proc CheckForPackage { name { version  "" } } {

# All packages needed for this script are available here...

	global tcl_platform
	
	set site "http://pages.infinit.net/cclients/software.htm"
	
# The following code assumes that UNIX users have netscape, Windows users have something.
# On other systems you have to roll your own code....

	if { [string length $version] != 0 } {
		set result [catch { package require $name $version } reason]
	} else {
		set result [catch { package require $name } reason]
		}
		
	if { $result } {
		puts stderr "Did not get package $name $version : $reason"
		puts stderr "This script requires the $name $version package!"
		puts stderr "Get it at $site"
		if { $tcl_platform(platform) == "unix" } {
			catch { exec $Data(Browser) $site & } 
		} else {
			catch { exec start $site & }
			}
		exit
		}
		
	return $reason
	}
	
# These are small utility packages that you can get from my web site

CheckForPackage TclUtils
CheckForPackage httpcopy

# Obviously, this script will not run without the Fltk package. If you are
# a keener, you can redo it on a Tk canvas, if you want.

CheckForPackage Fltk 0.4

# This is done because screen resolutions are different...

if { $tcl_platform(platform) == "unix" } {
	set Data(DefaultWidgets) 6	;# The number of widgets in the window
} else {
	set Data(DefaultWidgets) 4
	}

set Data(Scheme) "Image -f $Fltk(Library)/scripts/images/clouds.jpg"

# If you are on UNIX and have the Tclgdbm package you get a database of history data

if { [catch { package require Tclgdbm } reason] } {
	set Data(HasGdbm) 0
} else {
	set Data(HasGdbm) 1
	}

# This statement turns of http related download messages.

set Maps(Verbose) 0

# Set some default values. These stations are around Montreal, Canada, and
# Miami, just to make winter dreams more real....

set Data(Station) YUL,YMX,YHU,YOW,YGK,YSC,YYZ,YQB,KBOS,KJFK
#set Data(Station) CYUL

# This is the defaule site for the web server. It is an alias for several
# servers of different types.


set Data(BaseURL) www.weatheroffice.ec.gc.ca
set Data(Script,Canada) city/pages
set Data(Script,USA) forecast/city_e.html
set Data(Client,Canada) ?
set Data(Client,USA) ?
set Data(Token) {}
set Data(Browser) mozilla

if { $tcl_platform(platform) == "unix" } {
	set host [lindex [split [exec hostname] .] 0]
} else {
	set host windows
	}

# This code configures the script for a few of my machines. 
# You probably want to change this for your own situation. It sets
# up the location of the data storage area on my computers. Note that
# for Windows users, permissions on mapped drives may cause some problems
# with this script. You need write permission for the location of the
# temporary files used by the script.

switch -exact -- $host {

io		-
invicta		{
				set Data(DefaultRoot) /database/weather.data
				}
windows		{
				set Data(DefaultRoot) "c:"
				}
				
default		{	set Data(DefaultRoot) $env(HOME)
				}
			} ;# End of switch cases

# The log file collects information about what is happening. You may not need it.

set Data(LogFile) $Data(DefaultRoot)/station.log

# This is where image icons are stored. Once you have them all, the
# script will not download things from the web site. This makes things
# faster and less resource consuming, I presume.

set Data(Images) $Fltk(Library)/scripts/images/weather

# This change is just to centralize the collection of icons for preparation
# of distributions...

if { $host == "invicta" } {
	set Data(Images) /source/Fltk-0.3/scripts/images/weather
	}

set Data(DefaultMode) 0		;# Technical plot of the data
set Data(DefaultUnits) ""	;# Metric units please

if { $Data(HasGdbm) } {
	set Data(Database) $Data(DefaultRoot)/station.observations.gdbm
} else {
	}

set Data(MissingValue) "???"	;# Indicates a field is not present

set Data(GMT) 1			;# Times in GMT

set Data(HasNext) 0			;# State variable for history button
set Data(HasPrev) 0			;# State variable for history button

set Data(HistoryLimit) 24	;# Hours of history to keep in memory

# List of variables plotted in the time series charts

set VarList { Temperature Pressure DewPoint WindSpeed WindDirection}

# Display help information for this script

proc DisplayHelp {} {

	set Help "\nThe format of the command line is:\n\n \
			\tstnplot ?-a? ?scheme? ?-s? ?stations? ?-m? ?mode? ?-u? ?units?\n \
			where -s or -stations specifies a comma seperated list of stations to watch\n \
			\t\t?-v? ?verbose? ?-w? ?widgets? ?-h? ?history?\n \
			\t\t?-g? ?gmt? ?-d? ?database? ?-??\n\n \
			\t-m or -mode specifies the mode to display the data (plot or depiction)\n \
			\t-u or -units specifies the units to use (imperial or metric)\n \
			\t-v or -verbose specifies whether download information is displayed (yes or no)\n \
			\t-w or -widgets specifies the number of widgets to display (default 4)\n \
			\t-d or -database specifies the database to use for history\n \
			\t-g or -gmt specifies whether time is local or GMT\n \
			\t-h or -history specifies the number of hours of history to keep\n \
			\t-a or -appearance specifies the scheme to use\n \
			\t-?	or -help displays this help message.\n"

	puts $Help

	Exit
	}
	
# Parse the parameter list

foreach { option value } $argv {
	switch -exact -- [string tolower $option] {

"-a"			-
"-appearance"	{	set Data(Scheme) $value		}

"-d"			-
"-database"	{	set Data(Database) $value
				}
"-g"			-
"-gmt"		{	set Data(GMT) [IsEqual [string tolower $value] yes 1 0]
				}
"-m"			-
"-mode"		{	set Data(DefaultMode) [IsEqual [string tolower $value] plot 0 1]
				}	
"-s"			-
"-station"	{	set Data(Station) [string toupper $value] }

"-u"			-
"-units"		{	set Data(DefaultUnits) [IsEqual [string tolower $value] imperial "&Unit=I" ""]
				}
"-v"			-
"-verbose"	{	set Maps(Verbose) [IsEqual [string tolower $value] "yes"]
				}
"-w"			-
"-widgets"	{	set Data(DefaultWidgets) $value
				}
"-h"			-
"-history"	{	set Data(HistoryLimit) $value
				}
"-?"			-
"-help"		{	DisplayHelp
				}
				
default		{	puts "Option $option is not supported!"
				DisplayHelp
				}
			} ;# End of switch cases
		} ;# End of foreach loop

# The following command selects the scheme to use

eval "Scheme $Data(Scheme)"

# Translate the current sky cover to a name for a background image

set SkyConditions {

	{	"Mostly Cloudy"	"Cloudy"					6	}
	{	"Sunny"			"Sunny"					0	}
	{	"Clear"			"Clear"					0	}
	{	"Overcast"		"Cloudy"					8	}
	{  	"Mainly Clear"		"Cloudy_with_clear_periods"	2	}
	{  	"Mainly Sunny"		"Sunny_with_cloudy_periods"	2	}
	{	"Cloudy"			"Cloudy" 					4	}

	}

# Find sky condition

proc FindSkyCondition { station str hour } {

	global SkyConditions Data

	foreach entry $SkyConditions {
		if { [string first "[lindex $entry 0]" $str] != -1 } {
#			set Data($station,Image,$hour) $Data(Images)/[lindex $entry 1].bmp
			return [lindex $entry 2]
			}
		}

	return "-1"
	}

# Translate the obstructions to vision. Order is important in this list!
# It appears that this site uses the simple term, like "rain", to mean "light rain"
# and qualifies more intense events with the "moderate" and "heavy" modifiers.

set Obstructions {

	{	"light rainshower"	"RW-" 	"Showers"	}
	{	"moderate rainshower"	"RW"	"Rainy"	}
	{	"heavy rainshower"	"RW+"	"Rainy"	}
	{	"rainshower"		"RW-"	"Showers"	}
	{	"light snowshower"	"SW-"	"Cloudy_with_snow"	}
	{	"moderate snowshower"	"SW"	"Cloudy_with_snow"	}
	{	"heavy snowshower"	"SW+"	"Snowing"	}
	{	"snowshower"		"SW-"	"Cloudy_with_snow"	}
	{	"snow grains"		"SG"		"Cloudy_with_snow"	}
	{	"moderate freezing rain" "ZR" "Freezing_rain"	}
	{	"heavy freezing rain"	"ZR+" "Freezing_rain"	}
	{	"freezing rain"		"ZR-" "Freezing_rain"	}
	{	"light rain"		"R-"		"Rainy"	}
	{	"moderate rain"		"R"	"Rainy"	}
	{	"heavy rain"		"R+"		"Rainy"	}
	{	"rain"			"R-"		"Rainy"	}
	{	"precipitation"		"R-"	"Rainy"	}	
	{	"light snow"		"S-"		"Snowing"		}
	{	"moderate snow"		"S"	"Snowing"	}
	{	"heavy snow"		"S+"		"Snowing"		}
	{	"drifting snow"		"DS"	"Blowing_snow" }
	{	"blowing snow"		"BS"		"Blowing_snow" }
	{	"snow"			"S-"		"Snowing"	}
	{	"light drizzle"		"L-"	"Drizzle"	}
	{	"moderate drizzle"	"L"		"Drizzle"	}
	{	"freezing drizzle"	"ZL-"	"Freezing_rain"	}
	{	"drizzle"			"L-"		"drizzle"	}
	{	"heavy thunderstorm" 	"TRW+" "Thundershowers"	}
	{	"thunderstorm"		"TRW"	"Thundershowers"	}
	{	"thunder"			"T"		"Thundershowers"	}
	{	"haze"			"H"		"Haze"	}
	{	"smoke"			"K"		"Smoke"	}
	{	"mist"			"M"		"Mist"	}
	{	"fog"				"F"	"Foggy"	}
	{	"dust"			"D"		"Dust" }

	}

# Find all obstructions mentioned in a string

proc FindObstructions { station str hour } {

	global Obstructions Data

	set result ""
	
	set list [split $str ,]

	foreach item $list {
		foreach entry $Obstructions {
			if { [regexp -nocase "[lindex $entry 0]" $item] != 0 } {
				append result [lindex $entry 1]
				break
				}
			}
		}
	return $result
	}

# This is the list of things that can be shown by the station model

set Data(Names) [list "Temperature" "Pressure" "Visibility" "Humidity" \
			"DewPoint" "WindDirection" "WindSpeed" "WindGusts" \
			"WindChill" "Humidex" "Tendency" "CurrentWeather" "Image" \
			"TendencyCode" "MslPressure" "CloudAmount" "ObsTime" \
			"VisLocation" "MaxTemp" "MinTemp" "MeanTemp" "Sunrise" "Sunset" \
			"Moonrise" "Moonset" "Precip" "Icon" ]

# This variable is used to control the state of the station plot widget

set Data(Update) 1

# Save some data to a file

proc SaveDataToFile { what where } {

	set fd [open $where w]
	if { $fd != "" } {
		puts $fd $what
		close $fd
		}

	return $what
	}
	
# Look up the station page name from its identifier

set StationPageNames {

	"YHU"	"qc-109"
	"YMX"	"qc-123"
	"YTR"	"qc-130"
	"YQB"	"qc-133"
	"YSC"	"qc-136"
	"YUL"	"qc-147"
	"YOW"	"on-118"
	"YYZ"	"on-143"
	"YGK"	"on-69"
	"YHZ"	"ns-19"
	"YAW"	"ns-40"
	
	}
	
proc LookupStationPageName { name } {

	global StationPageNames
	
	if { [set ord [lsearch $StationPageNames $name]] != -1 } {
		return "[lindex $StationPageNames [incr ord]]_metric_e.html"
		}
	
	return ""
	}
# Make a request for a city

proc MakeQuery { name } {

	global Data

	set query "$Data(BaseURL)"

	switch -exact -- [string index $name 0] {

"K"		{
		set query "$query/$Data(Script,USA)$Data(Client,USA)$name"
		} 
"C"		{
		set name [string range $name 1 end]
		set query "$query/$Data(Script,Canada)/[LookupStationPageName $name]"
		}
default	{	
		set query "$query/$Data(Script,Canada)/[LookupStationPageName $name]"
		}
		} ;# End of switch cases

	return "$query$Data(DefaultUnits)"
	}

# Get the station data from the web site

proc LoadStationData { name } {

	global Data tcl_platform

	if { $tcl_platform(platform) == "unix" } {
		set Data(FileName) /tmp/$name[clock seconds].txt
	} else {
		set Data(FileName) $Data(DefaultRoot)/$name.txt
		}
		
	if { [catch { ::http::copy [MakeQuery $name] $Data(FileName) } reason] } {
		puts stderr $reason
		return 0
		}
		
	return 1
	}

# Get the observation data

proc GetObservedData { name } {

	set fd [open $name r]
	if { $fd == "" } {
		puts stderr "Failed to open file $name!"
		return ""
		}

	set txt [read $fd]

	close $fd
	
	return [CleanHTMLBlock $txt]
	}

# Strip the HTML tags from a block of data.This function also cleans
# up other stuff of no interest in the web page.

proc CleanHTMLBlock { txt } {

	regsub -all {(.*)Currently} $txt {} txt
#	regsub -all {These icons(.*)World Wide Web Site.} $txt {} txt
	regsub -all {<IMG+|<img+} $txt ":IMG" txt       ;# Image references
	regsub -all {<([^>])+>} $txt {} txt            ;# HTML Tags
	regsub -all -line {^([[:space:]])*\n} $txt {} txt 
	regsub -all {\t+} $txt { } txt                 ;# Tab characters
	regsub -all { +} $txt { } txt                  ;# Multiple blanks
	regsub -all -lineanchor {^( *)} $txt {} txt	  ;# Preceeding blanks
	regsub -all -line ^ $txt {} txt                ;# Empty lines
	regsub -all -line :\n^ $txt {: } txt		  ;# Data entries
	regsub -all -line {Temp(.?)\n^} $txt {Temp: } txt
	regsub -all -line {\n(.?)C } $txt {C} txt
	regsub -all -lineanchor {$\nat \n} $txt {at } txt   ;# Valid time
	regsub -all {&nbsp;} $txt { } txt
	
	return $txt
	}

# Get the numeric values from a string

set Data(ExtractNumbers) {([-0123456789.%])+}

proc GetNumberValue { str { width 4 } } {

	global Data

	if { ! [regexp -all $Data(ExtractNumbers) $str value] } {
		set value $Data(MissingValue)
		}

	set len [string length $value]

	while { $len < $width } {
		set value " $value"
		incr len
		}
		
	return $value
	}
	
# Make a time code from the observation time

proc MakeTimeCode { str } {

	return [clock scan $str -gmt 1]
	}
	
# Convert wind speed from KM/hr

proc ConvertWindSpeed { metric } {

	return $metric
	
#	return [expr $metric / 1.6]
	}

# Parse the wind direction and speed

proc ParseWind { station str hour } {

	global Data
			
	if { [regexp calm $str] } {
		set Data($station,WindSpeed,$hour) ""
		set Data($station,WindDirection,$hour) calm
		return
		}

	set dir [lindex $str 0]
	set Data($station,WindDirection,$hour) [FindDirection $dir]

	if { [string first gust $str] != -1 } {
			regexp -all $Data(ExtractNumbers) $str Data($station,WindSpeed,$hour) Data($station,WindGusts,$hour)
			set Data($station,WindSpeed,$hour) [ConvertWindSpeed $Data($station,WindSpeed,$hour)]
			set Data($station,WindGusts,$hour) [ConvertWindSpeed $Data($station,WindGusts,$hour)]
	} else {
			regexp -all $Data(ExtractNumbers) $str Data($station,WindSpeed,$hour)
			set Data($station,WindSpeed,$hour) [ConvertWindSpeed $Data($station,WindSpeed,$hour)]
		}
	}

# Pad out a value

proc Pad { value { len 3 } { with 0 } } {

	set length [string length $value]
	while { $length < $len } {
		set value $with$value
		incr length
		}

	return $value 
	}

# Convert a pressure value to millibars

proc Millibars { value } {

	global Data

	set value [string trimleft $value 0]

	if { $value != $Data(MissingValue) } {
		set value [expr int($value * 10)]
		return [expr $value % 1000]
	} else {
		return $value
		}
	}

# Parse the observation data

proc ParseData { station txt } {

	global Data
	
	set Data(ObsTime) $Data(MissingValue)
	
	set HasNormals [set Data([MakeKey $station HasNormals $Data(ObsTime)]) 0]
	set Data([MakeKey $station HasYesterday $Data(ObsTime)]) 0

	set idx 0
	set wx ""
	set list [split $txt \n]
	while { $idx < [llength $list] } {
		set line [lindex $list $idx]
		if { [string length $line] != 0 } {
			set line [string trim $line]
			switch -regexp -- $line {
		
"Observed at"		{	set parts [split $line :]
				set Data(ValidTime) "[lrange [split [lindex $parts 1]] end-3 end]:[lindex $parts 2]"
#				set wx [lindex $parts 3]
				set wx [lindex $list [incr idx]]
				set wx $wx[lindex $list [incr idx]]
				set Data(ObsTime) [MakeTimeCode $Data(ValidTime)]
				set Data($station,Image,$Data(ObsTime)) ""
				set Data($station,ObsTime,$Data(ObsTime)) $Data(ObsTime)
							}
"^Temperature"		{	if { [regsub {&deg;+} [lindex $list [incr idx]] {} tmp] != 0 } {
					set Data([MakeKey $station Temperature $Data(ObsTime)]) [GetNumberValue $tmp]
					}
				}

"^Pressure"		{	set tmp [FudgePressure [GetNumberValue [lindex $list [incr idx]]]]
				set Data([MakeKey $station MslPressure $Data(ObsTime)]) $tmp
				set Data([MakeKey $station Pressure $Data(ObsTime)]) [Pad [Millibars $tmp]]
				}
				
"Visibility"		{	set Data([MakeKey $station Visibility $Data(ObsTime)]) [GetNumberValue [lindex $list [incr idx]]]
				set Data([MakeKey $station VisLocation $Data(ObsTime)]) ""
				}
"^Dewpoint"		{	if { [regsub {&deg;+} [lindex $list [incr idx]] {} tmp] != 0 } {
					set Data([MakeKey $station DewPoint $Data(ObsTime)]) [GetNumberValue $tmp]
					}
				}
"^Humidity"		{	set Data([MakeKey $station Humidity $Data(ObsTime)]) [GetNumberValue [lindex $list [incr idx]]]
				}
"^Wind Chill"		{	set Data([MakeKey $station WindChill $Data(ObsTime)]) [GetNumberValue [lindex $list [incr idx]]]
				}
"^Wind"			{	ParseWind $station [lindex $list [incr idx]] $Data(ObsTime)
				}
"^Humidex"		{	set Data([MakeKey $station Humidex $Data(ObsTime)]) [GetNumberValue [lindex $list [incr idx]]]
				}
"^Regional Normals"	{	set HasNormals [set Data([MakeKey $station HasNormals $Data(ObsTime)]) 1]	
				set done 5
				while  { $done > 0  && $idx < [llength $list] } {
					set line [lindex $list [incr idx]]
					switch -regexp -- $line {
					
				"^Max Temp:"	{	if { [regsub {&deg;+} [string trim [lindex [split $line :] 1]] {} tmp] != 0 } {
								set Data([MakeKey $station MaxTemp $Data(ObsTime)]) [GetNumberValue $tmp]
								}				
						}
						
				"^Min Temp:"	{	if { [regsub {&deg;+} [string trim [lindex [split $line :] 1]] {} tmp] != 0 } {
								set Data([MakeKey $station MinTemp $Data(ObsTime)]) [GetNumberValue $tmp]				
								set done 0
								}
						}
						} 	;# End of switch cases
					if { $done > 0 } { incr done -1 }
					}	;# End of while loop
				}

"^Yesterday"		{	set Data([MakeKey $station HasYesterday $Data(ObsTime)]) 1
				set done 7
				while  { $done > 0  && $idx < [llength $list] } {
					set line [lindex $list [incr idx]]
					switch -regexp -- $line {
					
				"^Max Temp:"	{	if { [regsub {&deg;+} [string trim [lindex [split $line :] 1]] {} tmp] != 0 } {
								set Data([MakeKey $station MaxHier $Data(ObsTime)]) [GetNumberValue $tmp]
								}				
						}
						
				"^Min Temp:"	{	if { [regsub {&deg;+} [string trim [lindex [split $line :] 1]] {} tmp] != 0 } {
								set Data([MakeKey $station MinHier $Data(ObsTime)]) [GetNumberValue $tmp]
								}				
						}
						
				"^Precip Total"	{	set Data([MakeKey $station Precip $Data(ObsTime)]) [lindex $list [incr idx]]	
							set done 0
						}
						} 	;# End of switch cases
					if { $done > 0 } { incr done -1 }
					}	;# End of while loop
				}

"^Moonrise"		{	set Data([MakeKey $station Moonrise $Data(ObsTime)]) [lindex $list [incr idx]]
				}
"^Moonset"		{	set Data([MakeKey $station Moonset $Data(ObsTime)]) [lindex $list [incr idx]]
				}
"^Sunrise"		{	set Data([MakeKey $station Sunrise $Data(ObsTime)]) [lindex $list [incr idx]]
				}
"^Sunset"		{	set Data([MakeKey $station Sunset $Data(ObsTime)]) [lindex $list [incr idx]]
				}
"^5 Day Forecast"	{
				set key [MakeKey $station Forecasts $Data(ObsTime)]
				set Data($key) {}
				set done 40; while { $done > 0 && $idx < [llength $list] } {
					set line [string trim [lindex $list [incr idx]] ]
					switch -regexp -- $line {
					
			"Monday|Tuesday|Wednesday|Thursday|Friday|Saturday|Sunday"
				{	set period [string trim $line]
					lappend Data($key) $period
					}
					
			"IMG"	{	
#				set period [lindex [split $line :] 0]
#				lappend Data($key) $period
					if { [regexp {/weathericons/[0-9]+.gif} $line icon] == 0 } {
						set icon "Missing"
						}
					regexp {alt=\".+\"} $line weather
					regsub -all {\"|alt=} $weather {} weather
					set Data([MakeKey $station "$period,Icon" $Data(ObsTime)]) $icon
					set Data([MakeKey $station "$period,Wx" $Data(ObsTime)]) $weather
					set Data([MakeKey $station "$period,Low" $Data(ObsTime)]) "   "
					set Data([MakeKey $station "$period,High" $Data(ObsTime)]) "   "
					}
			"^Low"	{	regsub -all {&deg;|Low |C} [string trim $line] {} tmp
					set Data([MakeKey $station "$period,Low" $Data(ObsTime)]) $tmp 
					}
			"^High"	{	regsub -all {&deg;|High |C} [string trim $line] {} tmp
					set Data([MakeKey $station "$period,High" $Data(ObsTime)]) $tmp 
					}
			"These icons"	{	set done 0
					}
						} ;# End of switch cases
					if { $done > 0 } { incr done -1 }
					}	;# End of while loop
				}
default			{	if { [string length [string trim $line]] } {  } 
				}
						} ;#End of switch cases
				}
			incr idx	
			}

# If we did not get an observation time, then we probably can't get the data
# for some reason, probably because the station is not valid!
			
	if { $Data(ObsTime) == $Data(MissingValue) } {
		return 0
		}

	set Data($station,Latest) $Data(ObsTime)
	
	set Data([MakeKey $station Tendency $Data(ObsTime)]) $Data(MissingValue)

	ComputePressureTendency $station $Data(ObsTime)
	ComposeCurrentWeather $station $wx $Data(ObsTime)
	
	set Data([MakeKey $station CloudAmount $Data(ObsTime)]) [FindSkyCondition \
		$station $Data([MakeKey $station CurrentWeather $Data(ObsTime)]) $Data(ObsTime)]

	return 1
	}

# Get the tendency code from the change in pressure

proc GetTendencyCode { change } {

	if { $change < 0 } { 
		return 1 
	} else {
		if { $change > 0 } {
			return 3
		} else {
			return 2
			}
		}
	}

# Fake the pressure change

proc FudgePressure { p } {

	return $p
	
	if { $p == "???" } { return $p }
	return [expr $p + rand()/10.0]
	}

# Compute the pressure change between 2 times

proc PressureChange { station t1 t2 } {

	global Data

	set p1 [GetData $station MslPressure $t1]
	set p2 [GetData $station MslPressure $t2]

	if { $p1 == $Data(MissingValue) || $p2 == $Data(MissingValue) } {
		puts "Missing value for $station p1 = $p1 p2 = $p2"
		return 0
		}

	return [expr $p1 - $p2]
}
	
# Compute the pressure tendency

proc ComputePressureTendency { station hour } {

	global Data
	
	if { [HasDataForField $station Tendency $hour] } { return }

	set previous [expr $hour - (3 * 3600)]
	
	if { [HasDataForField $station MslPressure $hour] } {
		if { [HasDataForField $station MslPressure $previous] } {
			set change [PressureChange $station $hour $previous]
			set Data([MakeKey $station Tendency $hour]) [expr int($change * 100)]
		} else {
			puts "No history data for $station at [clock format $previous -gmt 1]"
			}

		set tm1 [expr $hour - 3600]
		set tm2 [expr $tm1 - 3600]

		if [HasDataForField $station MslPressure $tm2] {
			if [HasDataForField $station MslPressure $tm1] { 
				set change [PressureChange $station $tm1 $tm2]
				set key [MakeKey $station TendencyCode $hour]
				set Data($key) [GetTendencyCode $change]
				set change [PressureChange $station $hour $tm1]
				set Data($key) [expr ($Data($key) << 2) | [GetTendencyCode $change]]
				}
			} 
	} else {
		puts "Data for $station at [clock format $hour -gmt 1] is missing"
		}
	}

# Move a file

proc MoveFile { from to } {

	if { [catch { file copy $from $to } reason] } {
		puts "Failed to move file $from to $to because : $reason"
		catch { file rename $from [file tail $to] }
		}
	}
	
# Download an icon image. The round robin nature of how the CWS does
# its web serving is the reason for this rather bizzare code. You do
# not know which server will deal you the icon, and they use a mix
# of IIS and Apache with different file trees, and different icon sets!

# Image icons3/29.gif is the sky condition not available image

proc DownloadIconImage { icon name } {

	global Data Maps tcl_platform
	
	regsub -all { } $name {_} filename

	set filename $filename.gif

	set result 0

	if { $tcl_platform(platform) == "unix" } {
		set tmpfile tmp[clock seconds].gif
	} else {
		set tmpfile tmp.gif
		}
	
	if { ! [file exists $Data(Images)/$filename] } {
#		puts "Query $Data(BaseURL)/$icon"
		if { [catch { ::http::copy $Data(BaseURL)/$icon $tmpfile } reason] } {
			puts "Failed to get $icon because : $reason"
		} else {
			if { $Maps(IsImage) } {
				MoveFile $tmpfile $Data(Images)/$filename
				if { ! [file exists $Data(Images)/$icon] } {
					MoveFile $tmpfile $Data(Images)/$icon
					}
#				puts "Loaded $icon to $Data(Images)/$filename"
				set result 1
			} else {
				puts "$icon not found on server!"
				}
			}
		catch { file delete $tmpfile } reason	
	} else {
		set result 1
		}

	return $result	
	}
	
# Compose Current Weather

proc ComposeCurrentWeather { station txt hour } {


	global Data

	if { $txt != "" } {

# Get the icon name from the input data

#		puts $txt
#		set icon [lindex [split [lindex [split [lindex [split $txt] 2] >] 0] =] 1]
#		puts "Icon : $icon"
#		regsub -all {\"+} $icon {} tmp 
		regexp {/weathericons/[0-9]+.gif} $txt icon
		set Data([MakeKey $station Icon $hour]) [string range [string trim $icon] 1 end]

# Get the current weather description from the input data

		set Data([MakeKey $station CurrentWeather $hour]) [string trim [lindex [split $txt >] 1]]

		Log "Icon $Data($station,Icon,$hour) -> $Data($station,CurrentWeather,$hour)"
			
# See if we already have an image for this type of weather

		if { [DownloadIconImage $Data($station,Icon,$hour) $Data($station,CurrentWeather,$hour)] == 0 } {
			puts "No icon for $Data($station,CurrentWeather,$hour) at $station at $hour!"
			img $Data(Images)/weathericons/29.gif
		} else {
			set img $Data(Images)/$Data($station,Icon,$hour)
			}

		set Data($station,Image,$hour) $img

# Make space for the current weather symbols

			if { [GetData $station CurrentWeather $hour] != "" } {
				set len [expr [string length [GetData $station CurrentWeather $hour]] * 3]
				set Data([MakeKey $station WxLocation $hour]) -$len
				}
				
# Add in the visibility

			if { [GetData $station Visibility $hour] != $Data(MissingValue) } {
				set wx [FindObstructions $station [string tolower [GetData $station CurrentWeather $hour]] $hour]
				if { $Data([MakeKey $station Temperature $hour]) < -2 } {
					regsub -all R+ $wx S wx
					}
				if { $wx != "" } {
					set key [MakeKey $station Visibility $hour]
					set Data($key) "$Data($key)$wx"
					set Data([MakeKey $station VisLocation $hour]) "sp -[expr (4 * [string length $wx])] +0"
					}
				}
			}	
	}
	
# Draw the sky cover

proc DrawSkyCover { station hour } {

	global Data
	
# An arc drawn with fill active is a pie section of the current background
# color.

	set result "home bg white fill 1 "
	
	switch -exact -- $Data($station,CloudAmount,$hour) {
	
"0"			{	append result "bg blue arc 5 0 360 "
					}
"2"			{	append result "arc 5 0 90 " 
					}
"4"			{	append result "arc 5 0 180 "
					}
"6"			{	append result "arc 5 0 270 "
					}
"8"			{	append result "arc 5 0 360 "
					}
"-1"			{	append result "pc 40 arc 5 0 360 "
					}
				} ;# End of switch cases

	append result "( hm pc black sp -5 +0 rt 90 fd 10 hm sp +0 +5 fd 10 )"
 				
	return $result
	}
	
# Write a log file

proc Log { msg } {

	global Data
	
	set fd [open $Data(LogFile) a+]
	
	if { $fd != "" } {
		puts $fd "[clock format [clock seconds] ] : $msg"
		close $fd
		}
	}
	
# Log missing fields

proc LogMissingFields { station hour } {

	global Data
	
	set str ""
	foreach name $Data(Names) {
		if { [GetData $station $name $hour] == $Data(MissingValue) } {
			append str "$name "
			}
		}
		
	Log "Missing Items for $station $str"
	}

# Make a key for the database

proc MakeKey { station name hour } {

	return "$station,$name,$hour"
	}

# Check if some valid data is in the database

proc HasDataForField { station name hour } {

	global Data

	set key [MakeKey $station $name $hour]

	if [info exists Data($key)] {
		if { $Data($key) != $Data(MissingValue) } {
			return 1
			}
		}

	return 0
	}

# Save a data element to the database

proc SaveData { station name hour value } {

	global Data

	set key [MakeKey $station $name $hour]
	set Data($key) $value

	}

# Get some data from the database

proc GetData { station name hour } {

	global Data

	set key [MakeKey $station $name $hour]

	if [info exists Data($key)] {
		return $Data($key)
	} else {
		return $Data(MissingValue)
		}
	}
		
# Save a history of data for the station

proc StoreDataHistory { station hour } {

	global Data

	set record ""
	foreach name $Data(Names) {
		if { [GetData $station $name $hour] != $Data(MissingValue) } {
			if { [string length $record] != 0 } {
				append record "*$name=[GetData $station $name $hour]"
			} else {
				set record "$name=[GetData $station $name $hour]"
				}
			}
		}
	
	if { $Data(HasGdbm) } {
		if [file exists $Data(Database)] {
			set db [gdbm open $Data(Database) rw] 
		} else {	
			set db [gdbm open $Data(Database) rwc]
			}
		gdbm store $db "$station*$hour" "$record"

		gdbm close $db
	} else {
		}
	
	}

# Handle a background error

proc bgerror { msg } {

	global Data
	
	set Data(Token) [after [expr 30 * 60 * 1000] "UpdateStationList $Data(LastStationList)"]
	}
	
# Update the list of stations

proc UpdateStationList { stationlist } {

	global Data

# Deactivate the widget. No recursive clicking!

	set Data(Update) 0

# Cancel any pending updates

	if { $Data(Token) != "" } {
		after cancel $Data(Token)
		}

# Load the data for each of the specified stations

	foreach station [split $stationlist ,] {
		set windows [Windows list]
		set idx [lsearch -glob $windows *$station.plot]
		if { $idx != -1 } {
			if { [catch { UpdateStation [lindex [Windows list] $idx] $station } reason] } {
				puts "Failed to update station $station : $reason"
				}
		} else {
			puts "Can't find window for station $station in $windows!"
			}
		}
	
# Update the display for this station every 30 minutes at 15 minutes past the hour

	set now [clock seconds]
	set delay [expr (int(($now / 3600) * 3600 + 900)) - $now]
	
	if { $delay < 0 } { set delay [expr 30 * 60] }
	
#	puts "Delay is [expr $delay/60] minutes"
	
	set Data(LastStationList) $stationlist
	set Data(Token) [after [expr $delay * 1000] "UpdateStationList $stationlist"]

# All done. Activate the widget for input!

	set Data(Update) 1
	}
	
# Update the station data

proc UpdateStation { w station } {

	global Data Model

#	puts "UpdateStation $station"
					
# Get the latest data from the station	

	set Data($station,Update) 1

	if { [LoadStationData $station] } {
		set data [GetObservedData $Data(FileName)]
		SaveDataToFile $data "/tmp/$station.txt"
		if { [ParseData $station $data] == 0 } {
			Log "Data not available for $station"
			set Data($station,Model) $Model(NotAvailable)
			set Data(Update) 1
			set Data($station,Update) 0
			catch { file delete $Data(FileName) }
			return 1
			}
		StoreDataHistory $station $Data(ObsTime)
		catch { file delete $Data(FileName) }
	} else {
		Log "Failed to get data for $station!"
		set Data($station,Model) $Model(MissingData)
		set Data(Update) 1
		set Data($station,Update) 0
		catch { file delete $Data(FileName) }
		return 1
		}

	catch { file delete $Data(FileName) }

# Create the plotting model based on the received data and 
# display it in the window...

	if [catch { BuildPlotModel $w $station $Data(ObsTime) } reason] {
		puts "Failed to build the model for $station : $reason"
		}
	
# Keep a log of the missing fields

	LogMissingFields $station $Data(ObsTime)
	
# Point to the new data

	set l [GetParameter $w Label]
	
	SetParameter $l Hour $Data(ObsTime)

# Set the state of the navigation buttons

	SetButtonStates $station $Data(ObsTime)

# Manage memory

	CleanupMemory $station $Data(ObsTime)

	set Data($station,Update) 0
	}

# This is a macro to place an item

proc Item { station name hour where { opts "" } } {

	return "( sp -35 -25 bd 65 50 al $where $opts tx \"\$Data($station,$name,$hour)\" )"
	}			

# These turtle graphics scripts will draw a standard station plot style of 
# depiction of the weather data.

set Model(StationCircle) "cs ht home pc black ft helv ( pc red pd cr 6 )"
set Model(Temperature)	"\[Item \$station Temperature \$hour top,left\]"
set Model(Pressure) 	"\[Item \$station Pressure \$hour top,right\]"
set Model(DewPoint)		"\[Item \$station DewPoint \$hour bottom,left\]"
set Model(Visibility)	"( sp -35 +6 \$Data(\$station,VisLocation,\$hour) tx \"\$Data(\$station,Visibility,\$hour)\" )"
set Model(ObsTime)		"( al bottom,centered tx \$obstime )"
set Model(Humidity)		"\[Item \$station Humidity \$hour bottom,left { pc darkgreen }\]"
set Model(WindDirection)	"( pu pc black rt \$Data(\$station,WindDirection,\$hour) fd 6 pd fd 30"
set Model(WindChill)	"\[Item \$station WindChill \$hour bottom,right { pc blue }\]"
set Model(Humidex)		"\[Item \$station Humidex \$hour bottom,right { pc red }\]"
set Model(CurrentWeather) "\[BuildCurrentWeather \$Data(\$station,CurrentWeather,\$hour)\]"
set Model(WindGusts)	"( )"
set Model(Tendency)		"( sp +12 +6 tx \"\$Data(\$station,Tendency,\$hour)\" )"
set Model(Station) 		"( sp -12 -36 tx \"\$station\" )"
set Model(Image)		"( sp -30 -30 im \"\$Data(\$station,Image,\$hour)\" )"
set Model(MissingData)	"cs home ht pc red ( al centered tx \"Data Missing\" )"
set Model(NotAvailable)	"cs home ht pc red ( al centered tx \"Not Available\" )"
set Model(MissingImage) "cs home ht pc red ( al centered tx \"No Image\" )"
set Model(IconName)		"home ( sp -12 +35 tx \"\[GetIconName \$Data(\$station,Icon,\$hour)\]\" )"
set Model(MaxTemp)		" ( sp -40 -15 tx \"Maximum \$Data(\$station,MaxTemp,\$hour) \$degrees\" )"
set Model(MinTemp)		" ( sp -40 -0 tx \"Minimum \$Data(\$station,MinTemp,\$hour) \$degrees\" )"
set Model(MeanTemp)		" ( sp -40 +15 tx \"Mean \$Data(\$station,MeanTemp,\$hour) \$degrees\" )"
set Model(HierMaxTemp)		" ( sp -40 -15 tx \"Maximum \$Data(\$station,MaxHier,\$hour) \$degrees\" )"
set Model(HierMinTemp)		" ( sp -40 -0 tx \"Minimum \$Data(\$station,MinHier,\$hour) \$degrees\" )"
set Model(HierPrecip)		" ( sp -40 +15 tx \"Precip. \$Data(\$station,Precip,\$hour)\" )"
set Model(WindCondition)	"home ( sy +35 al center,x tx \"\[FindDirectionName \$Data(\$station,WindDirection,\$hour)\] \$Data(\$station,WindSpeed,\$hour)\" )"

# Pressure tendency codes 5 - falling, 6 - falling then steady, 
# 7 - falling then rising, 9 - steady then falling, 10 - steady, 
# 11 - steady then rising, 13 - rising then falling, 14 - rising then steady, 
# 15 - rising. 

set Model(5) "( hm sp +\$where -4 rt 160 fd 10 )"
set Model(6) "( hm sp +\$where -4 rt 160 fd 8 lt 70 fd 4 )"
set Model(7) "( hm sp +\$where -4 rt 160 fd 8 lt 120 fd 6 )"
set Model(9) "( hm sp +\$where -4 rt 90 fd 2 rt 70 fd 10 )"
set Model(10) "( hm sp +\$where +0 rt 90 fd 6 )"
set Model(11) "( hm sp +\$where +4 rt 90 fd 2 lt 70 fd 10 )"
set Model(13) "( hm sp +\$where +4 rt 20 fd 8 rt 140 fd 4 )"
set Model(14) "( hm sp +\$where +4 rt 20 fd 8 rt 70 fd 4 )"
set Model(15) "( hm sp +\$where +4 rt 20 fd 10 )"

# Set up the current weather description for the display model There is a limit
# or 19 characters available along 1 line on the display widget, so split longer
# descriptions to take 2 lines.

proc BuildCurrentWeather { str } {

	set result "( al top,centered tx "
	
	if { [string length $str] > 19 } {
		set list [split $str]
		if { [list length $list] < 2 } {
			append result "\"$str\""
		} else {
			append result "\"[lindex $list 0]\" al centered,x "
			append result "tx \"[lrange $list 1 end]\""
			}
	} else {
		append result "\"$str\""
		}
		
	append result "  )"
	
	return $result
	}

# Names of the wind directions and their wind directions in degrees...

set Data(Directions) {
	 N 0 North NNE 22.5 NorthNorthEast  NE 45 NorthEast ENE 67.5 EastNorthEast 
	 E 90 East ESE 112.5 EastSouthEast
	SE 135 SouthEast SSE 157.5 SouthSouthEast S 180 South SSW 202.5 SouthSouthWest 
	SW 225 SouthWest
	WSW 247.5 WestSouthWest W 270 West WNW 292.5 WestNorthWest NW 310 NorthWest 
	NNW 332.5 NorthNorthWest
	}

# Get the wind angle from a name

proc FindDirection { name } {

	global Data

	foreach { n s l } $Data(Directions) {
		if { $name == $n } { 
			return $s
			}
		}

	return 0
	}
	
# Get the direction name from the angle

proc FindDirectionName { angle } {

	global Data
	
	foreach { n s l } $Data(Directions) {
		if { $angle == $s } {
			return $l
			}
		}
		
	return $angle
	}

# Make the wind barbs according to the wind speed

proc MakeSpeed { speed } {

	global Data
	
	if { $speed == $Data(MissingValue) } { return "" }
	
	set speed [expr int($speed + 0.5)]

	set fast [expr int($speed / 50)]
	set full [expr int(($speed - ($fast * 50)) / 10)]
	set half [expr int(($speed % 10) / 5)]

# The command sequence for the wind barbs is appended to the  commands
# that draw the wind shaft. The barb placement is based on relative
# motion along the shaft.

	set str "lt 180"
	
	while { $fast } {
		append str " ( lt 105 fd 10 rt 177 fd 10 ) pu fd 4 pd"
		incr fast -1
		}

	while { $full } {
		append str " ( lt 105 fd 10 ) pu fd 4 pd"
		incr full -1
		}

	while { $half } {
		append str " ( lt 105 fd 6 ) pu fd 4 pd"
		incr half -1
		}

# This will pop the machine state back to where it was before the shaft
# was drawn.

	append str " )"

	return $str
	}
	
# Blend 2 values

proc Blend { a b factor } {

	return [expr int((1.0 - $factor) * $b + $factor * $a)]
	}
	
# Compute a color based on a temperature value

proc ComputeColor { temp } {

	set factor [expr ($temp + 50.0) / 100.0]
	
	return "[Blend 200 0 $factor],0,[Blend 0 200 $factor]"
	}
		
# Build the plot model based on the data available from the station

proc BuildPlotModel { w station hour } {

	global Data 
	global Model
	
#	puts "Building for $station at $hour"

	set obstime [clock format $hour -format "%d/%H[IsEqual $Data(GMT) 1 Z \"\"]" -gmt $Data(GMT)]
	
# The modes of the display window are ...
#
#	0 - Normal station plot window
#	1 - Iconic representation of the current weather
#	2 - Large font temperature display 
#	3 - Station normal values
#	4 - Yesterday's values
#	5-10 - Forecasts for the station
#
	switch $Data($station,Mode) {
	
5	-
6	-
7	-
8	-
9	-
10	{	set period [lindex $Data($station,Forecasts,$hour) [expr $Data($station,Mode) - 5]]
		set Data(Model) "cs ht"
		append Data(Model) " ( al top ft helv fs 10 tx \"$period\" )"
		append Data(Model) " ( sp -30 -30 im \"\$Data(Images)/\$Data(\$station,\$period,Icon,\$hour)\" )"
		append Data(Model) " ( home sy +35 al center,x tx  \"\$Data(\$station,\$period,Wx,\$hour)\" )"
		append Data(Model) " ( al bottom,right tx \"\$obstime\" )"
		append Data(Model) " ( al bottom,left tx \"\$Data(\$station,\$period,Low,\$hour)/\$Data(\$station,\$period,High,\$hour)\" )"		
		}


4	{	if { $Data(DefaultUnits) == "" } {
			set degrees "C"
		} else { 
			set degrees "F"
			}
		set Data(Model) "cs ht"
		append Data(Model) " ( al top ft helv fs 10 tx \"Yesterday\" )"
		if { $Data($station,HasYesterday,$hour) } {
			append Data(Model) $Model(HierMaxTemp)
			append Data(Model) $Model(HierMinTemp)
			append Data(Model) $Model(HierPrecip)
		} else {
			append Data(Model) " ( al centered tx \"Not available\" )"
			}
		append Data(Model) " ( al bottom ft helv fs 10 tx \"\$obstime\" )"
		}
3	{	if { $Data(DefaultUnits) == "" } {
			set degrees "C"
		} else { 
			set degrees "F"
			}
		set Data(Model) "cs ht"
		append Data(Model) " ( al top ft helv fs 10 tx \"Normals\" )"
		if { $Data($station,HasNormals,$hour) } {
			append Data(Model) $Model(MaxTemp)
			append Data(Model) $Model(MinTemp)
#			append Data(Model) $Model(MeanTemp)
		} else {
			append Data(Model) " ( al centered tx \"Not available\" )"
			}
		append Data(Model) " ( al bottom ft helv fs 10 tx \"\$obstime\" )"
		}
2	{	set Data(Model) "cs ht ( ft lucida,bold fs 50"
		set color [ComputeColor $Data($station,Temperature,$hour)]
		append Data(Model) " pc $color al centered tx \"\[string trim \$Data(\$station,Temperature,\$hour)\]\" )"
		append Data(Model) " $Model(CurrentWeather)"
		append Data(Model) " ( al bottom ft helv fs 10 tx \"\$obstime\" )"
		}

1	{
		set Data(Model) "cs ht"
		append Data(Model) " $Model(CurrentWeather)"
		append Data(Model) " $Model(Image)"
		append Data(Model) " $Model(WindCondition)"
		append Data(Model) " ( al bottom,right tx \"\$obstime\" )"
		append Data(Model) " ( al bottom,left tx \"\[string trim \$Data(\$station,Temperature,\$hour)\]/\[string trim \$Data(\$station,DewPoint,\$hour)\]\" )" 
		} 
0	{
	
		set Data(Model) $Model(StationCircle)
	
		foreach name $Data(Names) {
			if { [GetData $station $name $hour] != $Data(MissingValue) } {
				switch -exact -- $name {

"WindDirection"	{	append Data(Model) " $Model($name) [MakeSpeed [GetData $station WindSpeed $hour]]"
					}
"TendencyCode"		{	if { [set code $Data($station,TendencyCode,$hour)] != $Data(MissingValue) } {
						set where [expr [string length $Data($station,Tendency,$hour)] * 6 + 2 + 12]
						append Data(Model) " [subst $Model($code)]"
						}
					}
"Icon"			-
"MslPressure"		-
"Image"			-
"WindSpeed"		-
"CloudAmount"		-
"VisLocation"		-
"MaxTemp"			-
"MinTemp"			-
"MeanTemp"		-
"Precip"			-
"Sunrise"			-
"Moonrise"		-
"Sunset"			-
"Moonset"			{ }

"Humidity"		{	if { [GetData $station DewPoint $hour] == $Data(MissingValue) } {
						append Data(Model) " $Model($name)"
						}
					}
default			{	append Data(Model) " $Model($name)"
							}
					}	;# End of switch cases
				}
			}	;# End of foreach loop
		
			append Data(Model) " $Model(ObsTime) [DrawSkyCover $station $hour]"
			}
		}	;# Mode switch

	set Data($station,Pressure,$hour) [Pad [Millibars $Data($station,MslPressure,$hour)]]

#	puts "[subst [subst $Data(Model)]]"

	set Data($station,Model) "[subst [subst $Data(Model)]]"

# Handle problem situations. If the displayed observation time is more
# than 2 hours old when compared to the current time, the web site is not
# being updated, so show the observation on a red background.

	set since [expr [clock seconds] - $hour]

	if  { $since > 7200 } {
		if { $Data($station,HasNext) == 0  && $Data($station,Update) == 1 } {
			$w set -bg red
		} else {
			$w set -bg gray
			}
	} else {
		$w set -bg gray
		}
	}

# Handle a button press

proc HandleButtonPress { w button x y } {

	global Data VarList

	$w set -relief sunken -state disabled

	if { $button == 1 } {
		set station [GetParameter $w Station]
		if { [catch { UpdateStation $w $station } reason] } {
			puts "Can't update $station : $reason"
			}
	} else {
		catch { PlotTimeSeries [GetParameter $w Station] $VarList } 
#		catch { DisplayCurrentHistory [GetParameter $w Station] }
		}

	$w set -state 1 -relief raised

	}

# Get the name of an icon

proc GetIconName { name } {

	return [file tail $name]
	}

# Get a parameter from the label widget

proc GetParameter { w name } {

	set list [split [$w get -data] *]

	set idx [lsearch -glob $list $name=*]

	if { $idx == -1 } { return "" }

	return [lindex [split [lindex $list $idx] =] 1]
	}

# Set a parameter value in a the label widget

proc SetParameter { w name value } {

	set list [split [$w get -data] *]

	set idx [lsearch -glob $list $name=*]

	if { $idx == -1 } {
		lappend list "$name=$value"
		$w set -data [join $list *]
	} else {
		set list [lreplace $list $idx $idx "$name=$value"]
		$w set -data [join $list *]
		}

	}
	
# Set the mode of the station

proc SetStationMode { w mode } {

	global Data
	
	set station [GetParameter $w Station]
	set hour [GetParameter $w Hour]
	
	set Data($station,Mode) $mode

	foreach name $Data(Names) {
		set Data($name) [GetData $station $name $hour]
		}

	if { [catch { BuildPlotModel [GetParameter $w Drawing] $station $hour } reason] } {
		puts "BuildPlotModel : $reason"
		}
		
	return $mode
	}
	
# Query the web page for a station

proc DisplayWebPage { station } {

	global Data tcl_platform
	
	set query http://[MakeQuery $station]
	
	if { $tcl_platform(platform) == "windows" } {
		if { [catch { exec start $query } reason] } {
			puts "Failed to start $query because : $reason"
			}
	} else {
		if { [catch { exec $Data(Browser) $query & } reason] } {
			puts "Failed to start $query because : $reason"
			}
		}
	}
	
# Select a display mode

proc SelectDisplayMode { w station count hour x y } {

	global Data
	
	catch { rename m.popup "" ; Destroy m }
	
	set p [Popup m.popup]
	$p add command -label "Station Plot" -command "SetStationMode $w 0"
	$p add command -label "Weather Depiction" -command "SetStationMode $w 1"
	$p add command -label "Temperature Display" -command "SetStationMode $w 2"
	$p add command -label "Regional Normals" -command "SetStationMode $w 3"
	$p add command -label "Yesterday's Values" -command "SetStationMode $w 4"
	
	incr count -5
	if { $count > 0 } {
		$p add cascade -label "Forecast Data"
		for { set i 0 } { $i < $count } { incr i } {
			set period [lindex $Data($station,Forecasts,$hour) $i]
			$p add command -label $period -command "SetStationMode $w [expr $i + 5]"
			}
		$p add terminator
		}
		
	$p add command -label "Web Page" -command "DisplayWebPage $station"
		
	$p popup -x [expr $x + 50] -y $y
	
	}
	
# Handle a mode change for a station

proc HandleModeChange { w button x y } {

	global Data tcl_platform

	set station [GetParameter $w Station]
	set hour [GetParameter $w Hour]
	set count [expr 5 + [llength $Data($station,Forecasts,$hour)] ]

#	puts "$station mode is $Data($station,Mode) button $button"

		switch -exact -- $button {

	1	{
			SetStationMode $w [expr ($Data($station,Mode) + 1) % $count]
			} 

	2 	{	if { [catch { SelectDisplayMode $w $station $count $hour $x $y } reason] } {
				puts "SelectDisplayMode failed : $reason"
				}
			}
			
	3	{
			SetStationMode $w [expr ($Data($station,Mode) + $count - 1) % $count]
			}
		}	;# End of switch cases 
	}	;# HandleModeChange
	
# Check if there is any data for a specified hour

proc HasDataForHour { station hour } {

	global Data

	set hasdata 0
	foreach name $Data(Names) {
		if { [GetData $station $name $hour] != $Data(MissingValue) } {
			incr hasdata
			}
		}
		
	return [IsEqual $hasdata 0 0 1]
	}

# Show all previous observations

proc PreviousObservations {} {

	global Data

	foreach station [split $Data(Station) ,] {
		catch { PreviousObservation [Windows list $station,up] }
		}
	}

# Show all next observations

proc NextObservations {} {

	global Data

	foreach station [split $Data(Station) ,] {
		catch { NextObservation [Windows list $station,down] }
		}

	}

# Change the mode of all windows

proc ChangeModes {} {

	global Data

	foreach station [split $Data(Station) ,] {
		catch { HandleModeChange [Windows list $station,label] 20 20 }
		}		
	}
	
# Display the next observation for a station

proc NextObservation { w } {

	global Data

	set station [GetParameter $w Station]
	set label [GetParameter $w Label]
	
	set hour [expr [GetParameter $label Hour] + 3600]

#	puts "Loading observation for $station at [clock format $hour -gmt 1]"
			
	if { [HasDataForHour $station $hour] } {
		if { [catch { BuildPlotModel [GetParameter $w Drawing] $station $hour } reason] } {
			puts "Build : $reason"
			}
		SetParameter $label Hour $hour
		SetButtonStates $station $hour
	} else {
		puts "No next observation for $station at [clock format $hour -gmt $Data(GMT)]"
		}					
	
	}
	
# Display the previous observation for a station

proc PreviousObservation { w } {

	global Data
	
	set station [GetParameter $w Station]
	
	set label [GetParameter $w Label]
	
	if { [catch { set hour [expr [GetParameter $label Hour] - 3600] } reason] } {
		puts "Error : $reason"
		}
	
#	puts "Previous for $station is [clock format $hour -gmt 1] with label $label"

	if { [HasDataForHour $station $hour] } {
		if { [catch { BuildPlotModel [GetParameter $w Drawing] $station $hour } reason] } {
			puts "Build : $reason"
			}
		SetParameter $label Hour $hour
		SetButtonStates $station $hour
	} else {
		puts "$station has no previous data for [clock format $hour -gmt $Data(GMT)]"
		}					
	}
	
# Set the state variables for the history buttons

proc SetButtonStates { station hour } {

	global Data
	
	set Data($station,HasNext) [HasDataForHour $station [expr $hour + 3600]]
	set Data($station,HasPrev) [HasDataForHour $station [expr $hour - 3600]]

	set Data(HasNext) $Data($station,HasNext)
	set Data(HasPrev) $Data($station,HasPrev)
	}

# Filter a list

proc Filter { list station } {

	set result {}

	foreach item $list {
		if { [lindex [split $item *] 0] == $station } {
			lappend result $item
			}
		}

	return $result
	}

# Load the history from the database (Unix version only)

proc LoadFromDatabase { station } {

	global Data

	if { ! [file exists $Data(Database)] } {
		return
		}
	if { [catch { set db [gdbm open $Data(Database) r] } reason] } {
		puts "No history database available!"
		return
		}

	set list [Filter [gdbm list $db] $station]

	set now [clock seconds]
	
	foreach item $list {
		set parts [split $item *]
		set valid [lindex $parts 1]

		if { [expr $now - $valid] <= [expr $Data(HistoryLimit) * 3600] } {
			set record [gdbm fetch $db $item]
			foreach name $Data(Names) {
				set Data($station,$name,$valid) $Data(MissingValue)
				}
				
			set elements [split $record *]
			foreach element $elements {
				set l [split $element =]
				set Data($station,[lindex $l 0],$valid) [lindex $l 1]
				}
			}

		}

	gdbm close $db
	}
	
# Display the current history of observations

proc DisplayCurrentHistory { station } {

	global Data
	
	set list [lsort [array names Data $station,*]]
	
	catch { Destroy h }
	
	Listbox h.h -w 600 -h 200
	
	foreach ob $list {
		set l [split $ob ,]
		if { [llength $l] == 3 } {
			set hour [lindex $l 2]
			set name [lindex $l 1]
			set valid [clock format $hour -gmt 1]
			h.h add "$valid $name $Data($ob)"
			}
		}
		
	Show h
	
	Wm title h "Current history for $station"
	} 

# Cleanup old data. Here I remove anything older than the number of hours
# defined by the history limit value

proc CleanupMemory { station hour } {

	global Data

	set when [expr $hour - ($Data(HistoryLimit) * 3600)]

	foreach name $Data(Names) {
		set key [MakeKey $station $name $when]
		if { [info exists Data($key)] } {
			catch { unset Data($key) }
			}
		}
	}
	
# Compute a time index

proc TimeIndex { now then } {

	set hours [expr 24 - ($now - $then) / 3600]
	
	return $hours
	}

# Load the functions needed to time series plots
	
source [file dirname [info script]]/timesubs.tcl
							 	
# Create an interface to display the data

# Set up some highlight behaviour for the buttons

#Bind Button <Enter> { %W set -fg red }
#Bind Button <Leave> { %W set -fg black }

Bind Drawing <ButtonRelease> { %W set -relief raised }

# Make certain we begin from scratch

catch { Destroy t }

# Set up the Scroll if needed. If there are more stations in the list than
# will fit into the specified window, then wrap the list of widgets into
# a Scroll that has only a vertical scroll bar.

set stations [split $Data(Station) ,]
set count [llength $stations]
 
if { $count < $Data(DefaultWidgets) } {
	set f0 t
} else {
	set itemsize 120 ;# this is the label size plus the drawing size
	set windowsize [expr $Data(DefaultWidgets) * $itemsize]
	set f0 [Scroll t.scroll -ystep $itemsize -scrollbars vertical \
		-x 0 -y 0 -h $windowsize  -w 114]
	}

# Pack all the widgets into something that can be scrolled...

set f [Package $f0.all -o vertical -x 0 -y 0 -w 100]

if { $count > 1  && 1 == 0 } {
	set g [Package $f.global -orientation horizontal -w 100 -x 0 -y 0]
	RepeatButton $g.prev -text @2< -w 20 -command PreviousObservations \
		-tooltip "Click to see the previous observations" \
		-statevariable Data(HasPrev)
	Button $g.update -text @menu -command "UpdateStationList $Data(Station)" \
			-width 30 -tooltip "Click to update all stations" \
			-statevariable Data(Update)
	Button $g.mode -text @square -command ChangeModes \
			-width 30 -tooltip "Click to change display mode" \
			-statevariable Data(Update)
	RepeatButton $g.next -text @2> -w 20 -command NextObservations \
			-tooltip "Click to see the next observations" \
			-statevariable Data(HasNext)
	}

# Construct a widget for each station.

foreach s $stations {

	set Data($s,HasPrev) 0
	set Data($s,HasNext) 0

	if { $Data(HasGdbm) } {
		LoadFromDatabase $s
		}

	set Data($s,Mode) $Data(DefaultMode)
	set Data($s,Model) "cs"

# Each station is a label and a drawing widget packed into a mega-widget

	set f1 [Package $f.$s -w 100 -orientation vertical]

	set f2 [Package $f1.buttons -w 100 -orientation horizontal]
	
	RepeatButton $f2.up -text @2< -w 20 \
		-data "Station=$s*Drawing=$f1.plot*Label=$f2.label" \
		-tooltip "Display the previous observation for $s" \
		-statevariable Data($s,HasPrev) \
		-command { PreviousObservation %W }
	
	Button $f2.label -text $s -w 60 \
		-data "Station=$s*Drawing=$f1.plot" \
		-tooltip "Click to change mode of $s" \
		-command { HandleModeChange %W %b %x %y }

	RepeatButton $f2.down -text @2> -w 20 \
		-data "Station=$s*Drawing=$f1.plot*Label=$f2.label" \
		-tooltip "Display the next observation for $s" \
		-statevariable Data($s,HasNext) \
		-command { NextObservation %W }

	Drawing $f1.plot -relief raised -statevariable Data(Update) \
		-tooltip "Click to update the observation for $s" \
		-data "Station=$s*Label=$f2.label" -variable Data($s,Model)

	Bind $f1.plot <ButtonPress> { HandleButtonPress %W %b %x %y}
	}

# Display the resulting UI. Automatic window size is used to create a Toplevel
# widget that wraps the contents.
	
Show t

Wm title t "Stn V2.00"

# Get the data for the stations

UpdateStationList $Data(Station)


